{Copyright (c) John Wiley & Sons, Inc., 1996,97.  All rights reserved.
Reproduction or translation of any part of this work beyond that
permitted in Section 117 of the 1976 United States Copyright Act
without the express written permission of the copyright owner is
unlawful.  Requests for permission or further information should be
addressed to the Permissions Department, John Wiley & Sons, Inc.
605 Third Avenue, New York, New York 10158.  The purchaser may make
back-up copies for his/her own use only and not for distribution or
resale.  The Publisher assumes no responsibility for errors,
omissions, or damages, including, without limitation, damages caused
by the use of these programs or from the information contained herein.}

{************************************************
                      GULP
   Linear programming program in Turbo Pascal
************************************************}

{$R-}       {R- = Range checking off}
{$B+}       {B+ = Boolean complete evaluation on}
{$S+}       {S+ = Stack checking on}
{$I-}       {I- = I/O checking off}
{$N+,E+}    {N+ = numeric coprocessor on, E+ = emulation on}
{$V-}       {V- = strict type checking on string variables switched off}
{$DEFINE WILEY}

PROGRAM Gulp;

Uses Crt, Dos;

{These declarations  differ between versions}
CONST
  MaxNoCols = 200;
  PotentialNoRows = 200;
  MaxNoRows : INTEGER = PotentialNoRows;
  Version = 'GULP 4.42';
  ProgName = 'GULP';
  LastUpdate = '4 May 1997';
TYPE
  RealType = Extended;
  RowArrayType =  ARRAY [0..MaxNoCols] OF RealType;
  RowArrayPtr = ^RowArrayType;
VAR
  A : ARRAY [0..PotentialNoRows] OF RowArrayPtr;

{These don't}
CONST
  ScreenWidth = 80;
  ScreenHeight = 25;
  NameSize = 10;
  RHS : INTEGER = 0;
  ObjFn : INTEGER = 0;
  FF = ^L;
  HomeKey = #128;
  EndKey = #129;
  BarChar = #205;
  F1Key = #130;

TYPE
  ScreenString = STRING[ScreenWidth];
  NameType = STRING[NameSize];
  FileType = STRING[80];
  String20 = STRING[20];
  RowType = BYTE;
  ColType = BYTE;
  CharSetType = SET OF CHAR;
  ProblemRec = RECORD
    Name :NameType;
    Aim : STRING[3];
    Multiplier : -1..1;
    END;
  MethodType = (Highest, Steepest);
{TYPED} CONST
  ControlCardFile : FileType = '';
  BoundsFile : FileType = '';
  FileChars : CharSetType
      = [#33,#35..#41,#45,#48..#57,#64..#90,#94..#123,#125,#126];
  MethodStr : ARRAY[MethodType] OF STRING[8] = ('Highest','Steepest');
  ColorStr : ARRAY[0..15] OF STRING[13] =
    ('Black','Blue','Green','Cyan','Red','Magenta','Brown','Light gray',
     'Dark gray','Light blue','Light green','Light cyan','Light red',
     'Light magenta','Yellow','White');

VAR
  ColName : ARRAY [0..MaxNoCols]   OF NameType;
  RowName : ARRAY [0..PotentialNoRows] OF NameType;
  ConstraintType : ARRAY [0..PotentialNoRows] OF CHAR;
  ColThisAct, ActThisCol : ARRAY [-PotentialNoRows..MaxNoCols] OF INTEGER;
  OriginalObjFn : ARRAY [1..MaxNoCols] OF RealType;
  OriginalRHS : ARRAY [1..PotentialNoRows] OF RealType;
  ExitProcAddress : POINTER;
  JumpToEditor,DropIntoEditor,KeepSoln,Alterred,
    StopSolve,IncludeCols,IncludeRows,IncludeRange,ScreenLost,Aborted,
    FirstTime:BOOLEAN;
  Ch, Response : CHAR;
  Mode, Dest, InitString, ResetString : NameType;
  DirMask, WorkFile, InFile, OutFile, Ext : FileType;
  TopRow, BotRow, NoLs, NoEs, NoGs, NoRows, R, SR, FirstRow : RowType;
  TopCol, BotCol, NoCols, C, SC, FirstCol : ColType;
  Margin : 1..ScreenWidth;
  Problem : ProblemRec;
  ForegroundColor,
  BackgroundColor,
  ColWidth,Dec,I,It,PrinterWidth : INTEGER;
  InChan, OutChan : TEXT;
  DirString, GulpDir, S : ScreenString;
  RowsAvail : LONGINT;
  ThisColName, ThisRowName : NameType;
  HelpFileName,SLNFileName,FileStr : PathStr;
  ErrorCode : INTEGER;
  PrintPageStopped : BOOLEAN;
  Method1, Method2 : MethodType;

PROCEDURE Edit (VAR WorkFile : FileType); FORWARD;

{$I GULPBITS.PAS}

PROCEDURE Edit;

CONST
  MinColWidth = 5;
  MaxColsPerPage = (Screenwidth-14) DIV MinColWidth;
  RowsPerPage = ScreenHeight-3;
VAR
  SCPos : ARRAY [1..MaxColsPerPage] OF 1..ScreenWidth;
  SRPos : ARRAY [1..RowsPerPage]    OF 1..ScreenHeight;
  ColsPerPage : 1..MaxColsPerPage;
  FinishPage : BOOLEAN;
  PromptSet, MoveSet : CharSetType;
  Prompt : ScreenString;

PROCEDURE MoveCursor (Ch : CHAR); FORWARD;
PROCEDURE Commands; FORWARD;

PROCEDURE SetUpEditor;
  VAR
    SR : RowType; SC : ColType;
  BEGIN
  ColsPerPage := (ScreenWidth-14) DIV ColWidth;
  FOR SC := 1 TO ColsPerPage DO
    SCPos[SC] := (SC-1)*ColWidth + 14;
  FOR SR := 1 TO RowsPerPage DO
    SRPos[SR] := SR + 1;
  FinishPage := TRUE;
  Alterred := FALSE;
  MoveSet := [#9,^M,^E,^S,^X,^D,^H,^J,^K,^L,^A,^R,^F,^C,HomeKey,EndKey];
  PromptSet := ['/']; Ch := ' ';
  IF It = -1 THEN Prompt := 'Press [/] for commands'
  ELSE IF JumpToEditor THEN
    Prompt := 'Press [/] and then [C] to return to original matrix'
  ELSE IF DropIntoEditor THEN
    Prompt := 'Press [/] and then [C] to continue next iteration';
  FirstTime := FALSE;
  END {of setupeditor};

PROCEDURE GotoCell(Mode : NameType);
  BEGIN
  CASE Mode[1] OF
    'M'{atrix}   : GOTOXY(SCPos[SC],SRPos[SR]);
    'R'{ow name} : GOTOXY(1,SRPos[SR]);
    'C'{ol name} : GOTOXY(SCPos[SC],1);
    'r'{ow Type} : GOTOXY(12,SRPos[SR]);
    ELSE {mode = 'Prob name'} GOTOXY(1,1);
    END {of case};
  END {of gotocell};

PROCEDURE PrintCell(Mode : NameType);
  BEGIN
  CASE Mode[1] OF
    'M'{atrix}   : BEGIN
                 IF A[R]^[C] = 0 THEN S := '           '
                 ELSE STR(A[R]^[C]:1:Dec,S);
                 WRITE(COPY(S+'          ',1,ColWidth));
                 END;
    'R'{ow name} : BEGIN
      I := ActThisCol[1-FirstRow-SR];
      IF I <= 0 THEN WRITE(RowName[-I]:1)
      ELSE WRITE(ColName[I]:1);
      END;
    'C'{ol name} : BEGIN
      I := ActThisCol[SC + FirstCol - 1];
      IF I >= 0 THEN WRITE(COPY(ColName[I]+' ',1,ColWidth))
      ELSE WRITE(COPY(RowName[-I]+' ',1,ColWidth));
      END;
    'r'{ow Type} : BEGIN
                   GOTOXY(11,SRPos[SR]);
                   WRITE(' ',ConstraintType[R],' ');
                   END;
    ELSE {mode = 'Prob name'} WRITE(Problem.Name);
    END {of case};
  END {of printcell};

PROCEDURE OccupyCell;
  BEGIN
  IF Mode = 'Matrix' THEN LowVideo ELSE NormVideo;
  GotoCell('Row name'); PrintCell('Row name');
  GotoCell('Col name'); PrintCell('Col name');
  GotoCell(Mode);
  InverseVideo;
  PrintCell(Mode);
  GotoCell(Mode);
  NormVideo;
  END {of occupycell};

PROCEDURE PrintStatus(Pos : BYTE; S : FileType);
  BEGIN
  GOTOXY(Pos,ScreenHeight-1);
  InverseVideo;
  WRITE(S);
  NormVideo;
  END {of printstatus};

PROCEDURE PrintPage;

  PROCEDURE PrintNames;
    VAR SR, R, SC, C : INTEGER;
    BEGIN
    GOTOXY(1,1);
    WRITE(Problem.Name:1);
    SC := 0; SR := 0;
    WHILE (SC < ColsPerPage) AND NOT KEYPRESSED DO
      BEGIN
      SC := SC + 1;
      C := ActThisCol[SC + FirstCol - 1];
      GOTOXY(SCPos[SC],1);
      IF C >= 0 THEN WRITE(ColName[C]:1)
      ELSE WRITE(RowName[-C]:1);
      END {of for};
    WHILE (SR < RowsPerPage) AND NOT KEYPRESSED DO
      BEGIN
      SR := SR + 1;
      R := SR + FirstRow - 1;
      GOTOXY(12,SRPos[SR]);
      WRITE(ConstraintType[R]);
      GOTOXY(1,SRPos[SR]);
      R := ActThisCol[-R];
      IF R <= 0 THEN WRITE(RowName[-R]:1)
      ELSE WRITE(ColName[R]:1);
      END {of for};
    END {of printnames};

  PROCEDURE PrintCoeffs;
    VAR SR, R : RowType; SC, C : ColType;
    BEGIN
    SC := 0;
    WHILE (SC < ColsPerPage) AND NOT KEYPRESSED DO
      BEGIN
      SC := SC + 1;
      C:= SC + FirstCol - 1;
      SR := 0;
      WHILE (SR < RowsPerPage) AND NOT KEYPRESSED DO
        BEGIN
        SR := SR + 1;
        R := SR + FirstRow - 1;
        IF A[R]^[C] <> 0 THEN
          BEGIN
          GOTOXY(SCPos[SC],SRPos[SR]);
          WRITE(A[R]^[C]:1:Dec);
          END {of if};
        END {of for sr};
      END {of for sc};
    END {of printcoeffs};

  BEGIN {printpage}
  PrintPageStopped := TRUE;
  IF NOT KEYPRESSED THEN
    BEGIN
    CLRSCR;
    PrintNames;
    PrintCoeffs;
    IF NOT KEYPRESSED THEN
      BEGIN
      PrintStatus(1,Version);
      InverseVideo; WRITE(' ':ScreenWidth-LENGTH(Version));
      PrintStatus(13,COPY('Row='+IntToStr(R,0)+'  ',1,9));
      PrintStatus(22,COPY('Col='+IntToStr(C,0)+'  ',1,9));
      PrintStatus(31,'Aim='+Problem.Aim);
      S := Workfile;
      IF LENGTH(S) > 35 THEN
        S := COPY(WorkFile,LENGTH(S)-34,35);
      PrintStatus(41,'File='+S);
      Message (Prompt);
      PrintPageStopped := FALSE;
      END {of if};
    END {of if};
  END {of printpage};

PROCEDURE List(InFile : FileType);
  VAR
    EndOfPage : BOOLEAN;
    Line : ScreenString;
    J : BYTE;
    Ch : CHAR;
  BEGIN
  IF Exist(InFile) THEN
    BEGIN
    ASSIGN(InChan,InFile);
    RESET(InChan);
    IF IOError THEN EXIT;
    Ch := ' ';
    WHILE NOT EOF(InChan) AND (Ch <> #27) DO
      BEGIN
      J := 1;
      Clear;
      WRITELN;
      READLN(InChan,Line);
      REPEAT
        J := J + 1;
        WRITELN(Line);
        READLN(InChan,Line);
        EndOfPage := COPY(Line,1,1) = 'p';
      UNTIL EOF(InChan) OR (J = ScreenHeight-3) OR EndOfPage;
      If NOT EndOfPage THEN WRITELN(Line);
      GOTOXY(1,ScreenHeight);
      Centre('(Press [Esc] to exit or any other key to continue)');
      Ch := ReadKey;
      IF KEYPRESSED THEN Ch := ReadKey;
      END {of while};
    CLOSE(InChan);
    IF IOError THEN ;
    END {of list}
  ELSE
    BEGIN
    Message(InFile+' not found. Press any key.');
    Ch := ReadKey;
    END;
  END {of list};

PROCEDURE Help;
  BEGIN
  IF Exist(HelpFileName) THEN List(HelpFileName)
  ELSE ErrorMessage(1020);
  PrintPage;
  END {of help};

PROCEDURE MoveCell(VAR V,SV,FirstV:BYTE;V1,V2: BYTE; Increment : INTEGER;
  Mode1a,Mode1b,Mode1c,Mode2a,Mode2b,Mode2c : NameType; MovingRow : BOOLEAN);
  VAR
    N : NameType;
  BEGIN
  IF Mode = 'Matrix'THEN
    BEGIN
    IF MovingRow THEN N := 'Row name' ELSE N := 'Col name';
    GotoCell(N);
    PrintCell(N);
    END;
  GotoCell(Mode);
  PrintCell(Mode);
  IF (Mode[1] IN [Mode1a[1],Mode1b[1],Mode1c[1]]) AND
     ((Increment > 0) OR (Increment < 0) AND (V = 0)) THEN
    BEGIN
    IF Mode = Mode1a THEN Mode := Mode2a
    ELSE IF Mode = Mode1b THEN Mode := Mode2b
    ELSE IF Mode = Mode1c THEN Mode := Mode2c;
    END
  ELSE
    IF SV <> V1 THEN
      BEGIN
      SV := SV + Increment;
      V  := V  + Increment;
      IF MovingRow THEN PrintStatus(13,COPY('Row='+IntToStr(R,0)+'  ',1,9))
      ELSE PrintStatus(22,COPY('Col='+IntToStr(C,0)+'  ',1,9))
      END {of if}
    ELSE IF V <> V2 THEN
      BEGIN
      V := V + Increment;
      FirstV := FirstV + Increment;
      PrintPage;
      END {of else if};
  IF (V = V2) THEN FinishPage := TRUE ELSE FinishPage := FALSE;
  END {of movecell};

PROCEDURE MovePage(VAR V,FirstV : BYTE; Increment,MaxNoVs : INTEGER);
  VAR MaxFirstV, OldFirstV : BYTE; I : INTEGER;
  BEGIN
  MaxFirstV := MaxNoVs - Increment + 1;
  IF (0<FirstV) AND (Increment<0) OR (FirstV<MaxFirstV) AND (Increment>0) THEN
    BEGIN
    OldFirstV := FirstV;
    I := FirstV + Increment;
    IF I > MaxFirstV THEN FirstV := MaxNoVs - Increment + 1
    ELSE IF I < 0 THEN FirstV := 0
    ELSE FirstV := I;
    V := V + FirstV - OldFirstV;
    PrintPage;
    END {of if};
  END {of movepage};

PROCEDURE MoveToExtreme(VsPerPage,NoVs,NewV,NewSV : BYTE;
    VAR V,SV,FirstV : BYTE; MovingRow : BOOLEAN);
  VAR
    OldFirstV : BYTE;
    N : NameType;
  BEGIN
  OldFirstV := FirstV;
  IF Mode = 'Matrix'THEN
    BEGIN
    IF MovingRow THEN N := 'Row name' ELSE N := 'Col name';
    GotoCell(N);
    PrintCell(N);
    END;
  GotoCell(Mode);
  PrintCell(Mode);
  V := NewV;
  SV := NewSV;
  FirstV := 0;
  IF SV > 1 THEN
    BEGIN
    IF NoVs < VsPerPage THEN SV := NoVs + 1;
    I := NoVs - VsPerPage + 1;
    IF I > 0 THEN FirstV := I;
    END;
  IF OldFirstV = FirstV THEN
    IF MovingRow THEN PrintStatus(13,COPY('Row='+IntToStr(R,0)+'  ',1,9))
    ELSE PrintStatus(22,COPY('Col='+IntToStr(C,0)+'  ',1,9))
  ELSE PrintPage;
  END {of movetoextreme};

PROCEDURE DeleteOrInsert(DorI : CHAR);
  VAR
    Ch,ChN : CHAR;
    I,N : INTEGER;
    S1,S2 : STRING[7];

  PROCEDURE CopyVector(V1, V2:BYTE; VAR NoWs:BYTE; MovingRow:BOOLEAN);
    VAR V : BYTE;
    BEGIN
    FOR V := 0 TO NoWs DO
      IF MovingRow THEN A[V1]^[V] := A[V2]^[V]
      ELSE A[V]^[V1] := A[V]^[V2];
    IF MovingRow THEN
      BEGIN
      RowName[V1] := RowName[V2];
      IF V1 > ObjFn THEN ConstraintType[V1] := ConstraintType[V2];
      END
    ELSE ColName[V1] := ColName[V2];
    END {of copyvector};

  PROCEDURE DeleteOrInsertVector(VStart,VEnd : BYTE; VAR NoVs : BYTE;
         MaxNoVs,MaxNoWs : BYTE; Increment : INTEGER);
    VAR J,V,W : BYTE;
    BEGIN
    IF (VEnd - (VStart+(Increment-1)/2)) * Increment >= 0 THEN
      BEGIN
      ConstraintType[ObjFn] := 'L';
      I := 0;
      J := (Increment+1)DIV 2; {1=del,0=ins}
      WHILE (I < N) AND (NoVs < MaxNoVs) DO
        BEGIN
        V := VStart - I*(J-1);
        I := I + 1;
        WHILE V*Increment < (VEnd+J*(1-I))*Increment DO
          BEGIN
          V := V + Increment;
          CopyVector(V-Increment,V,MaxNoWs,(Ch='R'));
          END;
        FOR V := 0 TO MaxNoWs DO
          IF Ch = 'R' THEN A[VEnd+J*(1-I)]^[V] := 0
          ELSE A[V]^[VEnd+J*(1-I)] := 0;
        IF Increment < 0 THEN S1 := IntToStr(VEnd+N-I,0)
        ELSE S1 := IntToStr(NoVs,0);
        IF Ch = 'R' THEN
          BEGIN
          RowName[VEnd+J*(1-I)] := 'Row ' + S1 + '     ';
          ConstraintType[VEnd+J*(1-I)] := 'L';
          END
        ELSE ColName[VEnd+J*(1-I)] := 'Col ' + S1 + '     ';
        IF NoVs > 0 THEN NoVs := NoVs - Increment;
        END;
      ConstraintType[ObjFn] := 'N'; A[0]^[0] := 0;
      PrintPage;
      Alterred := TRUE;
      END {of if};
    END {of deleteorinsertvector};

  PROCEDURE Wipe;
    VAR
      Ch : CHAR;
    BEGIN
    IF It <> -1 THEN
      BEGIN
      Message('Return to original matrix before deleting all');
      Ch := ReadKey;
      END
    ELSE
      BEGIN
      Message('Delete whole matrix? (Y/N) : ');
      GetChar(Ch,['Y','N']);
      IF Ch = 'Y' THEN
        BEGIN
        Initialise(NoRows,NoCols);
        SetUpEditor;
        PrintPage;
        END;
      END;
    END {wipe};

  PROCEDURE DelBlock;
    VAR
      Row : RowType;
      Col : ColType;
    BEGIN
    IF (BotRow >= TopRow) AND (BotCol >= TopCol) THEN
      BEGIN
      FOR Row := TopRow TO BotRow DO
        FOR Col := TopCol TO BotCol DO A[Row]^[Col] := 0;
      Alterred := TRUE;
      PrintPage;
      END {of if}
    ELSE
      BEGIN
      Message('Block not properly defined. Reset top left &/or bottom right');
      Ch := ReadKey;
      END;
    END {delblock};


  BEGIN
  IF DorI = 'D' THEN
    Message('Delete - All, Rows, Columns or Block (A, R, C, B or [Esc]) : ')
  ELSE Message('Insert - Rows or Columns (R, C or [Esc]) : ');
  REPEAT Ch := ReadKey; Ch := UPCASE(Ch); ConvertKeyToIBM(Ch);
  UNTIL (Ch IN ['R','C',#27]) OR (Ch IN ['A','B']) AND (DorI = 'D');
  IF (Ch <> #27) THEN WRITE(Ch);
  IF Ch = 'A' THEN Wipe
  ELSE IF Ch = 'B' THEN DelBlock
  ELSE IF (Ch <> #27) THEN
    BEGIN
    IF Ch = 'R' THEN S1 := 'Rows' ELSE S1 := 'Columns';
    IF DorI = 'D' THEN S2 := 'Delete ' ELSE S2 := 'Insert ';
    Message(S2+S1+' - How many? (max 9) : ');
    GetChar(ChN,[#27,'0'..'9']);
    IF ChN <> #27 THEN
      BEGIN
      VAL(ChN,N,I);
      IF DorI = 'D'{elete} THEN
        IF Ch = 'R'{ows} THEN DeleteOrInsertVector(R,NoRows,NoRows,
            MaxNoRows+1,MaxNoCols,1)
        ELSE DeleteOrInsertVector(C,NoCols,NoCols,MaxNoCols+1,MaxNoRows,1)
      ELSE {Insert}
        IF Ch = 'R'{ows} THEN DeleteOrInsertVector(NoRows+1,R,NoRows,
            MaxNoRows,MaxNoCols,-1)
        ELSE DeleteOrInsertVector(NoCols+1,C,NoCols,MaxNoCols,MaxNoRows,-1);
      END {of if};
    END {of else};
  END {of DeleteOrInsert};

PROCEDURE CopyGulp;
  VAR Ch,ChN : CHAR; I,N : INTEGER; Row : RowType; Col : ColType;

  PROCEDURE CopyBlock;
    VAR
      Row,Col,FromRow,FromCol,ToRow,ToCol,DRow,DCol : INTEGER;
    PROCEDURE SetFromTo(V, MaxNoVecs : BYTE;
        TopVec,BotVec:INTEGER; VAR FromVec,ToVec,DVec : INTEGER);
      BEGIN
      IF V <= TopVec THEN
        BEGIN
        FromVec := V;
        ToVec := V + BotVec - TopVec;
        DVec := 1;
        END
      ELSE
        BEGIN
        FromVec := V + BotVec - TopVec;
        IF MaxNoVecs < FromVec THEN FromVec := MaxNoVecs;
        ToVec := V;
        DVec := -1;
        END;
      END {setfromto};
    BEGIN {copyblock}
    IF (BotRow >= TopRow) AND (BotCol >= TopCol) THEN
      BEGIN
      SetFromTo(R,MaxNoRows,TopRow,BotRow,FromRow,ToRow,DRow);
      SetFromTo(C,MaxNoCols,TopCol,BotCol,FromCol,ToCol,DCol);
      Row := FromRow;
      WHILE Row <> ToRow+DRow DO
        BEGIN
        Col := FromCol;
        WHILE Col <> ToCol+DCol DO
          BEGIN
          A[Row]^[Col] := A[Row-(R-TopRow)]^[Col-(C-TopCol)];
          IF A[Row]^[Col] <> 0 THEN
            BEGIN
            IF Row > NoRows THEN NoRows := Row;
            IF Col > NoCols THEN NoCols := Col;
            END;
          Col := Col + DCol;
          END;
        Row := Row + DRow;
        END;
     Alterred := TRUE;
      PrintPage;
      END {of if}
    ELSE
      BEGIN
      Message('Block not properly defined.');
      REPEAT I := I UNTIL KEYPRESSED;
      END;
    END {of copyblock};

  BEGIN
  Message('Copy - Row, Column or Block (R, C, B or [Esc]) : ');
  GetChar(Ch,['R','C','B',#27]);
  IF Ch <> #27 THEN
    BEGIN
    IF Ch = 'B' THEN CopyBlock
    ELSE IF Ch = 'R' THEN
      BEGIN
      I := R;
      Message('Enter row number to copy : ');
      READ(I);
      IF (I >= 0) AND (I <= MaxNoRows) THEN
        BEGIN
        FOR Col := 0 to NoCols DO
          A[R]^[Col] := A[I]^[Col];
        RowName[R] := RowName[I];
        ConstraintType[R] := ConstraintType[I];
        IF I = 0 THEN ConstraintType[R] := 'L';
        IF R > NoRows THEN NoRows := R;
        ConstraintType[ObjFn] := 'N';
        PrintPage;
        END {of if};
      END {of if}
    ELSE IF Ch = 'C' THEN
      BEGIN
      I := C;
      Message('Enter col number to copy : ');
      READ(I);
      IF (I >= 0) AND (I <= MaxNoCols) THEN
        BEGIN
        FOR Row := ObjFn to NoRows DO
          A[Row]^[C] := A[Row]^[I];
        ColName[C] := ColName[I];
        IF C > NoCols THEN NoCols := C;
        PrintPage;
        END {of if};
      END {of else if};
    Alterred := TRUE;
    END {of if upcase}
  ELSE Message(Prompt);
  A[ObjFn]^[RHS] := 0;
  END {of copy};

PROCEDURE PrintMatrix;
  VAR
    R, FirstRToPrint, LastRToPrint : RowType;
    C,FirstColThisPage, FirstCToPrint, LastCToPrint : ColType;
    I, J, Page : INTEGER;
    PCol, PCols : BYTE;
    Ch : CHAR;
    NumStr : STRING[12];
    Aborted : BOOLEAN;
    S : ScreenString;

  BEGIN {printmatrix}
  Aborted := FALSE;
  Message('Print matrix - All or Block (A, B or [Esc]) : ');
  GetChar(Ch,['A','B',#27]);
  IF Ch <> #27 THEN
    BEGIN
    IF Ch = 'A' THEN
      BEGIN
      FirstRToPrint := ObjFn;
      LastRToPrint := NoRows;
      FirstCToPrint := RHS;
      LastCToPrint := NoCols;
      END
    ELSE {ch = 'B'}
      IF (BotRow >= TopRow) AND (BotCol >= TopCol) THEN
        BEGIN
        FirstRToPrint := TopRow;
        LastRToPrint := BotRow;
        FirstCToPrint := TopCol;
        LastCToPrint := BotCol;
        END {of if}
      ELSE
        BEGIN
        Message('Block not properly defined. Reset top left &/or bottom right');
        Ch := ReadKey;
        Aborted := TRUE;
        END;
    IF NOT Aborted THEN AssignDest('MAT',Aborted);
    IF NOT Aborted THEN
      BEGIN
      IF Dest = 'Screen' THEN ScreenLost := TRUE
      ELSE Message('Writing matrix to ' + Dest);
      PCols := (PrinterWidth - 13) DIV (ColWidth + 1);
      IF Dest = 'Printer' THEN BEGIN
        CarefulWrite(Dest,OutChan,FALSE,InitString,ErrorCode);
        IF ErrorCode <> 0 THEN Exit;
        END;
      CarefulWrite(Dest,OutChan,TRUE,Version + ' ',ErrorCode);
      IF ErrorCode <> 0 THEN Exit;
      FirstColThisPage := FirstCToPrint;
      Page := 0;
      REPEAT
        Page := Page + 1;
        C := FirstColThisPage;
        CarefulWrite(Dest,OutChan,FALSE,'Page '+IntToStr(Page,0)+'   Problem:'+Problem.Name,ErrorCode);
        IF ErrorCode <> 0 THEN Exit;
        CarefulWrite(Dest,OutChan,TRUE,'',ErrorCode);
        IF ErrorCode <> 0 THEN Exit;
        CarefulWrite(Dest,OutChan,FALSE,'             ',ErrorCode);
        IF ErrorCode <> 0 THEN Exit;
        FOR PCol := 1 TO PCols DO
          BEGIN
          {find first blank in colname}
          I := NameSize;
          WHILE (I > 0) AND (ColName[C][I] = ' ') DO I := I - 1;
          IF I > ColWidth THEN I := ColWidth;
          S := Copy(ColName[C],1,I);
          IF C <= LastCToPrint THEN BEGIN
            CarefulWrite(Dest,OutChan,FALSE,PadStr(S,ColWidth)+' ',ErrorCode);
            IF ErrorCode <> 0 THEN Exit;
            END;
          C := C + 1;
          END {of for};
        CarefulWrite(Dest,OutChan,TRUE,'',ErrorCode);
        IF ErrorCode <> 0 THEN Exit;
        FOR R := FirstRToPrint TO LastRToPrint DO
          BEGIN
          C := FirstColThisPage;
          CarefulWrite(Dest,OutChan,FALSE,COPY(RowName[R]+'          ',1,NameSize)+
            ' '+ConstraintType[R]+' ',ErrorCode);
          IF ErrorCode <> 0 THEN Exit;
          FOR PCol := 1 TO PCols DO
            BEGIN
            IF C <= LastCToPrint THEN
              IF A[R]^[C] <> 0 THEN
                BEGIN
                STR(A[R]^[C]:ColWidth:Dec,NumStr);
                S := COPY(NumStr,1,ColWidth);
                CarefulWrite(Dest,OutChan,FALSE,PadStr(S,ColWidth)+' ',ErrorCode);
                IF ErrorCode <> 0 THEN Exit;
                END {of if}
              ELSE BEGIN
                CarefulWrite(Dest,OutChan,FALSE,PadStr(' ',ColWidth+1),ErrorCode);
                IF ErrorCode <> 0 THEN Exit;
                END;
            C := C + 1;
            END {of for pcol};
          CarefulWrite(Dest,OutChan,TRUE,'',ErrorCode);
          IF ErrorCode <> 0 THEN Exit;
          END {of for r};
        IF Dest = 'Printer' THEN BEGIN
          CarefulWrite(Dest,OutChan,FALSE,FF,ErrorCode);
          IF ErrorCode <> 0 THEN Exit;
          END
        ELSE BEGIN
          CarefulWrite(Dest,OutChan,TRUE,'',ErrorCode);
          IF ErrorCode <> 0 THEN Exit;
          END;
        FirstColThisPage := FirstColThisPage + PCols;
        IF Dest = 'Screen' THEN Ch := ReadKey;
      UNTIL C > LastCToPrint;
      IF Dest = 'Printer' THEN BEGIN
        CarefulWrite(Dest,OutChan,FALSE,ResetString,ErrorCode);
        IF ErrorCode <> 0 THEN Exit;
        END;
      CLOSE(OutChan);
      IF IOError THEN ;
      END;
    END;
  END {of printmatrix};

{OVERLAY} PROCEDURE EditName;

  PROCEDURE EditString (VAR S : NameType);
    CONST
      ColNo : BYTE = 44;
      Max : BYTE = 10;
    VAR
      X : INTEGER;
      OldS : FileType;
      Ch : CHAR;
      ChSet : CharSetType ;

    PROCEDURE GotoPos(X : INTEGER);
      BEGIN
      GOTOXY(X+ColNo-1,ScreenHeight);
      END;
    BEGIN
    Message('Edit : ');
    ChSet := [#32..#126];
    OldS := S;
    GotoPos(1);
    WRITE(S);
    GotoPos(1);
    X := 1;
    REPEAT
      Ch := ReadKey;
      ConvertKeyToIBM(Ch);
      CASE Ch OF
        #27 : S := OldS;
        ^L,^D : BEGIN {move right}
          IF (X <= LENGTH(S)) AND (X <= Max) THEN X := X + 1;
          GotoPos(X);
          END;
        #127,^G : IF X <= LENGTH(S) THEN {del char right}
          BEGIN
          DELETE(S,X,1);
          WRITE(COPY(S,X,LENGTH(S)-X+1),' ');
          GotoPos(X);
          END;
        ^S : BEGIN {move left}
          IF X > 1 THEN X := X - 1;
          GotoPos(X);
          END;
        ^F : BEGIN {far right}
          X := LENGTH(S) + 1;
          GotoPos(X);
          END;
        ^A : BEGIN {far left}
          X := 1;
          GotoPos(X);
          END;
        ^H : IF X > 1 THEN {del char left}
          BEGIN
          X := X-1;
          DELETE(S,X,1);
          GotoPos(1);
          WRITE(S,' ');
          GotoPos(X);
          END;
        ELSE IF Ch IN ChSet THEN
          BEGIN
          Insert(Ch,S,X);
          WRITE(COPY(S,X,LENGTH(S)-X+1),' ');
          IF X <= Max THEN X := X + 1;
          GotoPos(X);
          END;
        END {of case};
    UNTIL Ch IN [^M,#27];
    S := S + '          ';
    END {of editstring};

  BEGIN {editname}
  CASE Mode[1] OF
    'M'{atrix},'r'{ow Type} : BEGIN
      Message('Cannot edit this cell, only names.');
      REPEAT I := I UNTIL KEYPRESSED;
      END;
    'R' : EditString(RowName[R]);
    'C' : EditString(ColName[C]);
    'P' : EditString(Problem.name);
    END;
  END {of editname};

PROCEDURE DefineBlock;
  VAR
    Ch : CHAR;
  BEGIN
  Message('Mark block - Top left or Bottom right (T, B or [Esc]) : ');
  GetChar(Ch,['T','B',#27]);
  IF Ch = 'T' THEN
    BEGIN
    TopRow := R;
    TopCol := C;
    END
  ELSE IF Ch = 'B' THEN
    BEGIN
    BotRow := R;
    BotCol := C;
    END;
  Message(Prompt);
  END;

PROCEDURE Alter;
  CONST
    NoCoeffs = 17;
    Prompt = 'Enter your selection : ';
  TYPE
    String14 = STRING[14];
    CoeffType = RECORD
                Val : String14;
                ARPos, ACPos : INTEGER;
                END;
  VAR
    Coeff : ARRAY[1..NoCoeffs] OF CoeffType;
    Ch : CHAR;
    I, J : Integer;
    Aborted : BOOLEAN;

  FUNCTION Decimals (ASCII : NameType; MaxLength : BYTE) : ScreenString;
    VAR
      J : INTEGER; D : ScreenString;
    BEGIN
    D := '';
    FOR J := 1 TO LENGTH(ASCII) DO
      BEGIN
      STR(INTEGER(ASCII[J]),S);
      IF LENGTH(D) < MaxLength THEN D := D + S + ' ';
      END;
    Decimals := D;
    END {of decimals};

  FUNCTION YesNo (B : BOOLEAN) : String14;
    BEGIN
    IF B THEN YesNo := 'Yes' ELSE YesNo := 'No';
    END {of printval};

  PROCEDURE Convert(Ch : CHAR; VAR I : INTEGER);
    CONST ChArray : ARRAY[1..19] OF CHAR = 'NAWDOGCRPLIT12EFBSV';
    BEGIN
    I := 19;
    WHILE (I > 0) AND (Ch <> ChArray[I]) DO I := I - 1;
    END {of convert};

  PROCEDURE PrintLabels;
    VAR I : INTEGER;
    BEGIN
    Clear;
    WRITELN;WRITELN;
    Centre('Alter Default Settings');WRITELN;
    FOR I := 1 TO ScreenWidth DO WRITE(BarChar);
    GOTOXY(1,6);
    WRITELN('N : Name of data file                    P : Printer width');
    WRITELN('A : Aim of problem                       L : Logged directory');
    WRITELN('W : Width of column                      I : Init. string');
    WRITELN('D : Decimal places printed               T : reseT string');
    WRITELN('O : Output port                          1 : method phase 1');
    WRITELN('G : ranGe analysis?                      2 : method phase 2');
    WRITELN('C : Columns in output?                   E : drop into Editor?');
    WRITELN('R : Rows in output?                      F : Foreground color');
    WRITELN('S : Save these settings                  B : Background color');
    WRITELN('[Esc] : return to editor                 V : retrieVe defaults');
    FOR I := 1 TO ScreenWidth DO WRITE(BarChar);
    END {printlabels};

  PROCEDURE SetUpPositions;
    VAR I : INTEGER;
    BEGIN
    FOR I := 1 TO NoCoeffs DO WITH Coeff[I] DO
      IF I <= 8 THEN
        BEGIN
        ARPos := I + 5;
        ACPos := 28;
        END
      ELSE
        BEGIN
        ARPos := I - 3;
        ACPos := 64;
        END;
    END {setuppositions};

  PROCEDURE PrintVal(I : INTEGER);
    BEGIN
    IF I IN [1..NoCoeffs] THEN WITH Coeff[I] DO
      BEGIN
      CASE I OF
        1 : BEGIN
          S := Workfile;
          IF LENGTH(S) > 13 THEN
            Val := COPY(WorkFile,LENGTH(S)-12,13)
          ELSE Val := WorkFile;
          END;
        2 : Val := Problem.Aim;
        3 : STR(ColWidth,Val);
        4 : STR(Dec,Val);
        5 : Val := Dest;
        6 : Val := YesNo(IncludeRange);
        7 : Val := YesNo(IncludeCols);
        8 : Val := YesNo(IncludeRows);
        9 : STR(PrinterWidth,Val);
        10: Val := DirString;
        11: Val := Decimals(InitString,14);
        12: Val := Decimals(ResetString,14);
        13: Val := MethodStr[Method1];
        14: Val := MethodStr[Method2];
        15: Val := YesNo(DropIntoEditor);
        16: Val := ColorStr[ForegroundColor];
        17: Val := ColorStr[BackgroundColor];
        END {of case};
      GOTOXY(ACPos,ARPos);
      WRITE(' ':14);
      GOTOXY(ACPos,ARPos);
      WRITE(Val);
      END {of with};
    END {of printval};

  PROCEDURE GetInt(Msg :ScreenString; VAR Int : INTEGER; Min,Max : INTEGER);
    VAR
      S : STRING[4];
      OldInt,Err : INTEGER;
      Ch : CHAR;
    BEGIN
    OldInt := Int;
    REPEAT
      S := '';
      Int := OldInt;
      REPEAT
        Message(Msg+S);
        Ch := ReadKey;
        ConvertKeyToIBM(Ch);
        IF Ch IN ['0'..'9'] THEN S := S + Ch
        ELSE IF (Ch IN [#127,^S,^H]) AND (S<>'') THEN DELETE(S,LENGTH(S),1);
      UNTIL Ch IN [^M,#27];
      IF Ch = #27 THEN S := '';
      IF S <> '' THEN VAL(S,Int,Err);
      GOTOXY(1,ScreenHeight-1);
      IF (Int > Max) OR (Int < Min) THEN
        Centre('Value out of range. Re-enter.')
      ELSE CLREOL;
      IF S = '' THEN Int := OldInt;
    UNTIL (Int <= Max) AND (Int >= Min);
    END {of getint};

  PROCEDURE GetStr(Msg : ScreenString; VAR Str : NameType);
    VAR
      Ch : CHAR;
    BEGIN
    Str := '';
    REPEAT
      GOTOXY(1,25);
      WRITE(Msg+Decimals(Str,56));
      CLREOL;
      Ch := ReadKey;
      ConvertKeyToIBM(Ch);
      IF Ch <> ^M THEN
        IF Ch <> #8 THEN Str := Str + Ch
        ELSE DELETE(Str,LENGTH(Str),1);
    UNTIL Ch = ^M;
    END {of getstr};

  PROCEDURE SaveDefaults;
    BEGIN
    OutFile := GulpDir+ProgName+'.CFG';
    ASSIGN(OutChan,OutFile);
    REWRITE(OutChan);
    IF IOError THEN Exit;
    CarefulWrite('File',OutChan,TRUE,Mode,ErrorCode);
    IF ErrorCode <> 0 THEN Exit;
    CarefulWrite('File',OutChan,TRUE,WorkFile,ErrorCode);
    IF ErrorCode <> 0 THEN Exit;
    CarefulWrite('File',OutChan,TRUE,Problem.Aim,ErrorCode);
    IF ErrorCode <> 0 THEN Exit;
    CarefulWrite('File',OutChan,TRUE,IntToStr(ColWidth,0),ErrorCode);
    IF ErrorCode <> 0 THEN Exit;
    CarefulWrite('File',OutChan,TRUE,IntToStr(Dec,0),ErrorCode);
    IF ErrorCode <> 0 THEN Exit;
    CarefulWrite('File',OutChan,TRUE,Dest,ErrorCode);
    IF ErrorCode <> 0 THEN Exit;
    CarefulWrite('File',OutChan,TRUE,YesNo(IncludeCols),ErrorCode);
    IF ErrorCode <> 0 THEN Exit;
    CarefulWrite('File',OutChan,TRUE,YesNo(IncludeRows),ErrorCode);
    IF ErrorCode <> 0 THEN Exit;
    CarefulWrite('File',OutChan,TRUE,YesNo(IncludeRange),ErrorCode);
    IF ErrorCode <> 0 THEN Exit;
    CarefulWrite('File',OutChan,TRUE,IntToStr(PrinterWidth,0),ErrorCode);
    IF ErrorCode <> 0 THEN Exit;
    CarefulWrite('File',OutChan,TRUE,DirString,ErrorCode);
    IF ErrorCode <> 0 THEN Exit;
    CarefulWrite('File',OutChan,TRUE,Decimals(InitString,80),ErrorCode);
    IF ErrorCode <> 0 THEN Exit;
    CarefulWrite('File',OutChan,TRUE,Decimals(ResetString,80),ErrorCode);
    IF ErrorCode <> 0 THEN Exit;
    CarefulWrite('File',OutChan,TRUE,MethodStr[Method1],ErrorCode);
    IF ErrorCode <> 0 THEN Exit;
    CarefulWrite('File',OutChan,TRUE,MethodStr[Method2],ErrorCode);
    IF ErrorCode <> 0 THEN Exit;
    CarefulWrite('File',OutChan,TRUE,YesNo(DropIntoEditor),ErrorCode);
    IF ErrorCode <> 0 THEN Exit;
    CarefulWrite('File',OutChan,TRUE,IntToStr(ForegroundColor,0),ErrorCode);
    IF ErrorCode <> 0 THEN Exit;
    CarefulWrite('File',OutChan,TRUE,IntToStr(BackgroundColor,0),ErrorCode);
    IF ErrorCode <> 0 THEN Exit;
    CLOSE(OutChan);
    IF IOError THEN ;
    END {savedefaults};

  PROCEDURE ChangeVal(I : INTEGER);
    VAR
      Ch1 : CHAR;
      N : Byte;
      J : INTEGER;
      S : ScreenString;
    BEGIN {changeval}
    WITH Coeff[I] DO
      BEGIN
      CASE I OF
      1 : BEGIN
        GOTOXY(1,25);CLREOL;
        WRITE('Enter file : ');
        ReadFileName(WorkFile,'GLP',14,Aborted);
        END;
      2 : WITH Problem DO
        BEGIN
        IF Aim = 'MAX' THEN Aim := 'MIN' ELSE Aim := 'MAX';
        Multiplier := -1*Multiplier;
        END;
      3 : BEGIN
        GetInt('Enter column width (min 5, max 11) : ',ColWidth,5,11);
        ColsPerPage := (ScreenWidth - 14) DIV ColWidth;
        FOR J := 1 TO ColsPerPage DO
          SCPos[J] := (J-1)*ColWidth + 14;
        IF FirstCol + ColsPerPage - 1 > MaxNoCols THEN
          BEGIN
          FirstCol := MaxNoCols - ColsPerPage + 1;
          SC := 1;
          C := FirstCol;
          END {of if}
        ELSE IF C >= FirstCol + ColsPerPage THEN
          BEGIN
          SC := ColsPerPage;
          C := FirstCol + ColsPerPage - 1;
          END;
        END {of 3};
      4 : GetInt('Enter decimal places (min 0, max 9) : ',Dec,0,9);
      5 : IF Dest = 'Screen' THEN Dest := 'Printer'
        ELSE IF Dest = 'Printer' THEN Dest := 'File'
        ELSE Dest := 'Screen';
      6 : IncludeRange := NOT IncludeRange;
      7 : IncludeCols := NOT IncludeCols;
      8 : IncludeRows := NOT IncludeRows;
      9 : GetInt('Enter printer width (min 20, max 250) : ',PrinterWidth,20,250);
      10: BEGIN
        GOTOXY(1,25);
        WRITE('Drive:\Subdirectory : ');
        S := '';
        REPEAT
          GOTOXY(23,25);
          WRITE(S);
          CLREOL;
          GetChar(Ch,FileChars+[#13,#8,^S,':','\','.']);
          IF Ch <> ^M THEN
            IF (Ch IN FileChars+[':','\','.']) THEN S := S + Ch
            ELSE IF (Ch = #8) OR (Ch = ^S) THEN DELETE(S,LENGTH(S),1);
        UNTIL Ch = ^M;
        IF S <> '' THEN DirString := S;
        UpperCase(DirString);
        {$I-}
        N := LENGTH(DirString);
        IF (DirString[N] = '\') AND (N > 1) THEN
          IF (DirString[N-1] <> ':') THEN DELETE(DirString,N,1);
        IF (DirString[2] = ':') AND (N = 2) THEN BEGIN
          N := INTEGER(DirString[1]) - 64;
          GetDir(N,DirString);
          IF IORESULT <> 0 THEN DirString := '.';
          END;
        ChDir(DirString);
        IF IORESULT <> 0 THEN BEGIN
          Message('Subdirectory not found - Press any key.');
          Ch1 := READKEY;
          END;
        { $I+}
        GetDir(0,DirString);
        if DirString[length(DirString)] <> '\' then DirString := DirString + '\';
        END;
      11: GetStr('Enter initialisation string : ',InitString);
      12: GetStr('Enter reset string : ',ResetString);
      13: IF Method1 = Highest THEN Method1 := Steepest
          ELSE Method1 := Highest;
      14: IF Method2 = Highest THEN Method2 := Steepest
          ELSE Method2 := Highest;
      15: DropIntoEditor := NOT DropIntoEditor;
      16: BEGIN
        GOTOXY(1,20);
        Centre('0:Black 1:Blue 2:Green 3:Cyan 4:Red 5:Magenta 6:Brown 7:Light gray');
        GOTOXY(1,21);
        Centre('8:Dark gray 9:Light blue 10:Light green 11:Light cyan 12:Light red');
        GOTOXY(1,22);
        Centre('13:Light magenta 14:Yellow 15:White');
        GetInt('Enter foreground color (min 0, max 15) : ',ForegroundColor,0,15);
        NormVideo;
        PrintLabels;
        FOR J := 1 TO NoCoeffs DO PrintVal(J);
        END;
      17: BEGIN
        GOTOXY(1,20);
        Centre('0:Black 1:Blue 2:Green 3:Cyan 4:Red 5:Magenta 6:Brown 7:Light gray');
        GetInt('Enter background color (min 0, max 7) : ',BackgroundColor,0,7);
        NormVideo;
        PrintLabels;
        FOR J := 1 TO NoCoeffs DO PrintVal(J);
        END;
      18: BEGIN Message('Saving defaults'); SaveDefaults; END;
      19: BEGIN
        ReadDefaults;
        PrintLabels;
        FOR J := 1 TO NoCoeffs DO PrintVal(J);
        END;
      END {of case i};
      END {of with coeff[i]};
    PrintVal(I);
    Message(Prompt);
    END {of changeval};

  BEGIN
  PrintLabels;
  SetUpPositions;
  FOR I := 1 TO NoCoeffs DO PrintVal(I);
  Message(Prompt);
  REPEAT
    GetChar(Ch,[#27,'A'..'W','1','2']);
    Convert(Ch,I);
    ChangeVal(I);
  UNTIL Ch = #27;
  PrintPage;
  END {of alter};

PROCEDURE FileCommands (CallSave : BOOLEAN);
  VAR
    Ch, ChYN, Ch1 : CHAR;
    N : BYTE;
    FName : FileType;
    F : TEXT;
    S : ScreenString;
    DirOn, Aborted : BOOLEAN;
  PROCEDURE Save (VAR Workfile : FileType);
    VAR
      S : ScreenString; OutChan : TEXT;
      R : RowType; C : ColType;
      OrigName,BackupFile,TempFile : FileType;
    PROCEDURE WriteRows (C : ColType);
      VAR
        PosNo : 1..2;
        Coeff : String[20];
        R : INTEGER;
      BEGIN
      PosNo := 1;
      FOR R := ObjFn TO NoRows DO
        IF A[R]^[C] <> 0 THEN
          BEGIN
          STR(A[R]^[C]:12:10,Coeff);
          Coeff := COPY(Coeff,1,12);
          IF PosNo = 1 THEN
            BEGIN
            CarefulWrite('File',OutChan,FALSE,'    '+FillStr(ColName[C],10)+FillStr(RowName[R],10)+Coeff,ErrorCode);
            IF ErrorCode <> 0 THEN Exit;
            PosNo := 2;
            END {of if}
          ELSE
            BEGIN
            CarefulWrite('File',OutChan,FALSE,'   '+FillStr(RowName[R],10)+Coeff,ErrorCode);
            IF ErrorCode <> 0 THEN Exit;
            PosNo := 1;
            END {of else};
          IF PosNo = 1 THEN BEGIN
            CarefulWrite('File',OutChan,TRUE,'',ErrorCode);
            IF ErrorCode <> 0 THEN Exit;
            END;
          END {of if and for};
        IF PosNo = 2 THEN BEGIN
          CarefulWrite('File',OutChan,TRUE,'',ErrorCode);
          IF ErrorCode <> 0 THEN Exit;
          END;
      END {of WriteRows};

    PROCEDURE CopyControlCards;
      VAR
        F : TEXT;
      BEGIN
      IF (ControlCardFile <> '') THEN IF (Exist(ControlCardFile)) THEN BEGIN
        ASSIGN(F,ControlCardFile);
        RESET(F);
        IF IOError THEN EXIT;
        WHILE NOT EOF(F) DO BEGIN
          READLN(F,S);
          CarefulWrite('File',OutChan,TRUE,S,ErrorCode);
          IF ErrorCode <> 0 THEN Exit;
          END;
        CLOSE(F);
        IF IOError THEN ;
        END;
      END;

    PROCEDURE CopyBounds;
      VAR
        F : TEXT;
      BEGIN
      IF (BoundsFile <> '') THEN IF (Exist(BoundsFile)) THEN BEGIN
        ASSIGN(F,BoundsFile);
        RESET(F);
        IF IOError THEN EXIT;
        WHILE NOT EOF(F) DO BEGIN
          READLN(F,S);
          CarefulWrite('File',OutChan,TRUE,S,ErrorCode);
          IF ErrorCode <> 0 THEN Exit;
          END;
        CLOSE(F);
        IF IOError THEN ;
        END;
      END;

    BEGIN {save}
    Alterred := FALSE;
    GOTOXY(1,25);CLREOL;
    WRITE('Save - Enter file : ');
    ReadFileName(WorkFile,'GLP',21,Aborted);
    IF NOT Aborted THEN
      BEGIN
      Message('Saving ' + WorkFile);
      OrigName := WorkFile;
      WHILE OrigName[LENGTH(OrigName)] <> '.' DO DELETE(OrigName,LENGTH(OrigName),1);
      TempFile := OrigName + 'TMP';
      BackupFile := OrigName + 'BAK';
      ASSIGN(OutChan,TempFile);
      {$I-}
      REWRITE(OutChan);
      IF IORESULT <> 0 THEN BEGIN
        Message('Error: drive or subdirectory does not exist - Press any key.');
        Ch1 := ReadKey;
        IF KEYPRESSED THEN Ch1 := READKEY;
        END
      { $I+}
      ELSE BEGIN
        CopyControlCards;
        IF ErrorCode <> 0 THEN Exit;
        CarefulWrite('File',OutChan,TRUE,'NAME          '+Problem.Name,ErrorCode);
        IF ErrorCode <> 0 THEN Exit;
        CarefulWrite('File',OutChan,TRUE,'ROWS',ErrorCode);
        IF ErrorCode <> 0 THEN Exit;
        FOR R := ObjFn TO NoRows DO BEGIN
          CarefulWrite('File',OutChan,TRUE,' '+ConstraintType[R]+'  '+FillStr(RowName[R],10),ErrorCode);
          IF ErrorCode <> 0 THEN Exit;
          END;
        CarefulWrite('File',OutChan,TRUE,'COLUMNS',ErrorCode);
        IF ErrorCode <> 0 THEN Exit;
        FOR C := 1 TO NoCols DO WriteRows(C);
        CarefulWrite('File',OutChan,TRUE,'RHS',ErrorCode);
        IF ErrorCode <> 0 THEN Exit;
        WriteRows(RHS);
        IF ErrorCode <> 0 THEN Exit;
        CopyBounds;
        IF ErrorCode <> 0 THEN Exit;
        CarefulWrite('File',OutChan,TRUE,'ENDATA',ErrorCode);
        IF ErrorCode <> 0 THEN Exit;
        CLOSE(OutChan);
        IF IOError THEN Exit;
        IF Exist(WorkFile) THEN
          BEGIN
          IF Exist(BackupFile) THEN
            BEGIN
            ASSIGN(F,BackupFile);
            ERASE(F);
            IF IOError THEN Exit;
            END;
          ASSIGN(F,WorkFile);
          RENAME(F,BackupFile);
          IF IOError THEN Exit;
          END {of if};
        RENAME(OutChan,WorkFile);
        IF IOError THEN Exit;
        IF NOT DirOn THEN BEGIN
          S := WorkFile;
          IF LENGTH(S) > 35 THEN
            S := COPY(WorkFile,LENGTH(S)-34,35)
          ELSE WHILE LENGTH(S) < 35 DO S := S + ' ';
          PrintStatus(41,'File='+S);
          END;
        END {else};
      END {of if};
    END {of save};

  BEGIN {filecommands}
  IF CallSave THEN Save(WorkFile)
  ELSE
    BEGIN
    DirOn := FALSE;
    FName := WorkFile;
    REPEAT
      IF DirOn THEN Directory(DirMask);
      Message('File commands - Save,Retrieve,Dir,Erase,reName,List,Change dir,[Esc] ');
      GetChar(Ch,[#27,'S','R','D','E','N','L','C']);
      CASE Ch OF
        'S' : Save(WorkFile);
        'R' : BEGIN
          IF It <> -1 THEN
            BEGIN
            Message('Return to original matrix before retrieving next file');
            Ch := ReadKey;
            END
          ELSE
            BEGIN
            IF Alterred THEN
              BEGIN
              Message('Retrieve - Save previous Work-File? (Y/N) : ');
              GetChar(ChYN,['Y','N']);
              IF ChYN = 'Y' THEN Save(WorkFile);
              END {of if};
            IF NOT DirOn THEN Directory(DirString+'*.GLP');
            DirOn := TRUE;
            GOTOXY(1,25);CLREOL;
            WRITE('Retrieve - Enter file : ');
            ReadFileName(WorkFile,'GLP',25,Aborted);
            IF NOT Aborted THEN
              BEGIN
              Message('Deleting previous file');
              Initialise(NoRows,NoCols);
              SetUpEditor;
              Alterred := FALSE;
              IF Exist(WorkFile) THEN
                BEGIN
                Message('Retrieving '+WorkFile);
                Load(WorkFile);
                END {of if}
              ELSE BEGIN
                {Check if subdirectory exists}
                S := WorkFile;
                WHILE NOT (S[LENGTH(S)] IN ['\',':']) AND (LENGTH(S) > 0) DO
                  DELETE(S,LENGTH(S),1);
                IF S[LENGTH(S)] = '\' THEN DELETE(S,LENGTH(S),1);
                IF LENGTH(S) >0 THEN BEGIN
                  {$I-}
                  ChDir(S);
                  IF IORESULT <> 0 THEN BEGIN
                    Message('Subdirectory does not exist - Press any key.');
                    Ch1 := READKEY;
                    IF KEYPRESSED THEN Ch1 := READKEY;
                    WorkFile := COPY(WorkFile,LENGTH(S)+1,LENGTH(WorkFile)-LENGTH(S));
                    END;
                  { $I+}
                  END;
                Message('New file');
                END {else};
              END {if not aborted};
            END {else};
          END {of 'R'};
        'C' : BEGIN
          GOTOXY(1,25);CLREOL;
          WRITE('Enter Drive:\Subdirectory : ');
          ReadFileName(DirString,'',29,Aborted);
          IF DirString[LENGTH(DirString)] = '.' THEN
            DELETE(DirString,LENGTH(DirString),1);
          {$I-}
          N := LENGTH(DirString);
          IF (DirString[N] = '\') AND (N > 1) THEN
            IF (DirString[N-1] <> ':') THEN DELETE(DirString,N,1);
          IF (DirString[2] = ':') AND (N = 2) THEN BEGIN
            N := INTEGER(DirString[1]) - 64;
            GetDir(N,DirString);
            IF IORESULT <> 0 THEN DirString := '.';
            END;
          ChDir(DirString);
          IF IORESULT <> 0 THEN BEGIN
            Message('Subdirectory '+DirString+' not found - Press any key.');
            Ch1 := READKEY;
            END;
          { $I+}
          GetDir(0,DirString);
          if DirString[length(DirString)] <> '\' then DirString := DirString + '\';
          N := LENGTH(DirMask);
          WHILE (DirMask[N]<>'\') AND (DirMask[N]<>':') AND (N>0) DO N := N-1;
          DirMask := DirString + COPY(DirMask,N+1,12);
          END;
        'D' : BEGIN
          DirOn := TRUE;
          GOTOXY(1,25);CLREOL;
          WRITE('Enter dir mask : ');
          ReadFileName(DirMask,'GLP',18,Aborted);
          N := POS('\.',DirMask);
          IF N = 0 THEN N := POS(':.',DirMask);
          IF N <> 0 THEN INSERT('*',DirMask,N+1);
          END;
        'E' : BEGIN
          GOTOXY(1,25);CLREOL;
          WRITE('Erase - enter file : ');
          ReadFileName(FName,'GLP',22,Aborted);
          IF NOT Aborted THEN
            BEGIN
            IF Exist(FName) THEN
              BEGIN
              ASSIGN(F,FName);
              ERASE(F);
              IF IOError THEN ;
              END
            ELSE
              BEGIN
              Message(FName+' not found. Press any key.');
              Ch := ReadKey;
              END;
            END;
          END;
        'N' : BEGIN
          GOTOXY(1,25);CLREOL;
          WRITE('Rename - enter file : ');
          ReadFileName(FName,'GLP',23,Aborted);
          IF NOT Aborted THEN
            BEGIN
            IF Exist(FName) THEN
              BEGIN
              ASSIGN(F,FName);
              GOTOXY(1,25);CLREOL;
              WRITE('Enter new name : ');
              ReadFileName(FName,'',18,Aborted);
              IF NOT Aborted THEN
                BEGIN
                IF NOT Exist(FName) THEN BEGIN
                  RENAME(F,FName);
                  IF IOError THEN ;
                  END
                ELSE
                  BEGIN
                  Message(FName+' already exists. Request ignored. Press any key.');
                  Ch := ReadKey;
                  END;
                END;
              END
            ELSE
              BEGIN
              Message(FName+' not found. Press any key.');
              Ch := ReadKey;
              END;
            END;
          END;
        'L' : BEGIN
          GOTOXY(1,25);CLREOL;
          WRITE('List a file - enter file : ');
          ReadFileName(FName,'GLP',28,Aborted);
          IF (COPY(FName,LENGTH(FName)-2,3) <> 'COM') AND NOT Aborted AND
          (COPY(FName,LENGTH(FName)-2,3) <> 'EXE') THEN
            List(FName);
          IF NOT DirOn AND NOT Aborted THEN PrintPage;
          END;
        END {of case};
    UNTIL Ch IN [#27,'S','R'];
    IF DirOn OR (Ch = 'R') AND (NOT Aborted) THEN PrintPage
    ELSE Message(Prompt);
    END {of else};
  END {filecommands};

PROCEDURE GetCell (Ch {first character of string} : CHAR);
  TYPE String12 = STRING[12];
       ChSetType = SET OF ' '..#127;
  VAR
    Err : INTEGER;
    S : String12;
    X : RealType;
    Cha : CHAR;
  PROCEDURE GetString (VAR S: String12; Len : INTEGER; ChSet : ChSetType);
    BEGIN
    S := Ch; {i.e. the character becomes the first letter in the string}
    GOTOXY(1,ScreenHeight);
    WRITE(' Enter : ');
    REPEAT
      GOTOXY(10,ScreenHeight);
      WRITE(S);
      CLREOL;
      REPEAT
        Cha := READKEY;
        ConvertKeyToIBM(Cha);
      UNTIL (Cha IN ChSet + MoveSet + [#27]);
      IF (Cha = #127) OR (Cha = ^H) THEN DELETE(S,LENGTH(S),1)
      ELSE IF Cha = #27 THEN S := ''
      ELSE IF NOT (Cha IN MoveSet)
              AND (LENGTH(S) < Len) THEN S := S + Cha;
    UNTIL (Cha IN MoveSet-[^H]) OR (LENGTH(S) = 0);
    Message(Prompt);
    END {of getstring};

  BEGIN {getcell}
  S := '';
  Cha := ' ';
  IF Mode = 'Matrix' THEN
    IF ((R > 0) OR (C > 0)) AND (Ch IN ['0'..'9','-','.']) THEN
      BEGIN
      S := '';
      GetString(S,12,[^M,#127,'.','-','0'..'9']);
      IF S[1] = '.' THEN S := '0'+S
      ELSE IF COPY(S,1,2) = '-.' THEN INSERT('0',S,2);
      IF S[LENGTH(S)] = '.' THEN S := S+'0';
      VAL(S,X,Err);
      IF (Err = 0) AND (LENGTH(S) > 0) THEN
        BEGIN
        A[R]^[C] := X;
        IF R > NoRows THEN NoRows := R;
        IF C > NoCols THEN NoCols := C;
        Alterred := TRUE;
        END {of if}
      ELSE IF Err <> 0 THEN
        BEGIN
        Message('Not a valid number. Ignored.');
        Cha := READKEY;
        Message(Prompt);
        END;
      END {of if}
    ELSE {r = 0, c = 0 or ch not in set}
  ELSE IF Mode[1] IN ['P'{rob name},'C'{ol name},'R'{ow name}] THEN
    BEGIN
    GetString(S,10,[^M,' '..#127]);
    IF LENGTH(S) > 0 THEN
      BEGIN
      S := S + '         ';
      CASE Mode[1] OF
        'R'{ow name} : BEGIN RowName[R] := S;
                       IF R > NoRows THEN NoRows := R; END;
        'C'{ol name} : BEGIN ColName[C] := S;
                       IF C > NoCols THEN NoCols := C; END;
        'P'{rob name} : Problem.Name := S;
        END {of case};
      Alterred := TRUE;
      END {of if};
    END {of else}
  ELSE {mode = [type]}
    BEGIN
    Ch := UPCASE(Ch);
    IF R > 0 THEN
      IF Ch IN ['L','E','G'] THEN
        BEGIN
        ConstraintType[R] := Ch;
        IF Ch = 'G' THEN NoGs := NoGs + 1;
        IF R > NoRows THEN NoRows := R;
        Alterred := TRUE;
        END {of if};
    END {of else};
  OccupyCell;
  IF Cha IN MoveSet THEN MoveCursor(Cha);
  END {of getcell};

PROCEDURE FinishEdit;
  VAR
    ChYN : CHAR;
  BEGIN
  IF NoCols = 0 THEN
    BEGIN
    Message('Before selecting this option, enter some data or retrieve a data file.');
    Ch := ReadKey;
    Ch := ' ';
    ChYN := #27;
    END
  ELSE IF Alterred AND (It = -1) THEN
    BEGIN
    Message('Calculate - data not saved. Save? (Y/N) : ');
    GetChar(ChYN,['Y','N',#27]);
    IF ChYN = 'Y' THEN FileCommands(TRUE)
    ELSE IF ChYN = #27 THEN
      BEGIN
      Message(Prompt);
      Ch := ' ';
      END;
    END {of if};
  IF (ChYN <> #27) AND (It = -1) THEN Message('Calculating solution')
  ELSE IF Ch = ' ' THEN Message(Prompt);
  END {of finishedit};

PROCEDURE Quit;
  BEGIN
  IF Alterred THEN BEGIN
    Message('Quit - Data not saved. Leave GULP anyway? (Y/N) : ');
    GetChar(Ch,['Y','N']);
    END
  ELSE Ch := 'Y';
  IF Ch = 'Y' THEN
    BEGIN
    Clear;
    WRITELN;
    IF (ControlCardFile <> '') THEN IF (Exist(ControlCardFile)) THEN BEGIN
      ASSIGN(OutChan,ControlCardFile);
      ERASE(OutChan);
      IF IOError THEN ;
      END;
    IF (BoundsFile <> '') THEN IF (Exist(BoundsFile)) THEN BEGIN
      ASSIGN(OutChan,BoundsFile);
      ERASE(OutChan);
      IF IOError THEN ;
      END;
    HALT;
    END
  ELSE Message(Prompt);
  END {of quit};

PROCEDURE Print;
  VAR Ch,Ch1 : CHAR;
    OriginalDest : NameType;
  BEGIN
  ScreenLost := FALSE;
  OriginalDest := Dest;
  Message('Print - to Screen, Printer or File? (S,P,F or [Esc]) : ');
  GetChar(Ch,['S','P','F',#27]);
  CASE Ch OF
    'S' : Dest := 'Screen';
    'P' : Dest := 'Printer';
    'F' : Dest := 'File';
    END;
  IF Ch <> #27 THEN
    BEGIN
    Message('Print - Matrix, Solution or nothing? (M,S or [Esc]) : ');
    GetChar(Ch,['M','S',#27]);
    IF Ch = 'M' THEN PrintMatrix
    ELSE
      IF Ch = 'S' THEN
        IF It = -1 THEN PrintSoln(FALSE)
        ELSE
          BEGIN
          Message('Return to original matrix before printing last output.');
          Ch1 := READKEY;
          END;
    END;
  Dest := OriginalDest;
  IF ScreenLost THEN PrintPage ELSE Message(Prompt);
  END {of print};

PROCEDURE MoveCursor; {parameter in forward declaration at start of edit}
  BEGIN
  CASE Ch OF
    ^E,^K : MoveCell(R,SR,FirstRow,1,0,-1,'Matrix','row Type','Row name',
            'Col name','Prob name','Prob name',TRUE);
    ^S,^H : MoveCell(C,SC,FirstCol,1,0,-1,'Matrix','row Type','Col name',
            'row Type','Row name','Prob name',FALSE);
    ^X,^J : MoveCell(R,SR,FirstRow,RowsPerPage,MaxNoRows,1,'Col name',
            'Prob name','Prob name','Matrix','Row name','Row name',TRUE);
    ^D,^L : MoveCell(C,SC,FirstCol,ColsPerPage,MaxNoCols,1,'row Type',
            'Prob name','Row name','Matrix','Col name','row Type',FALSE);
    ^R : MovePage(R,FirstRow,-RowsPerPage,MaxNoRows);
    ^C : MovePage(R,FirstRow, RowsPerPage,MaxNoRows);
    ^A : MovePage(C,FirstCol,-ColsPerPage,MaxNoCols);
    #9,^F : MovePage(C,FirstCol, ColsPerPage,MaxNoCols);
    HomeKey : BEGIN
      MoveToExtreme(RowsPerPage,NoRows,0,1,R,SR,FirstRow,TRUE);
      MoveToExtreme(ColsPerPage,NoCols,0,1,C,SC,FirstCol,FALSE);
      END;
    EndKey : BEGIN
      MoveToExtreme(RowsPerPage,NoRows,NoRows,RowsPerPage,R,SR,FirstRow,TRUE);
      MoveToExtreme(ColsPerPage,NoCols,NoCols,ColsPerPage,C,SC,FirstCol,FALSE);
      END;
    '/' : Commands;
    END {of case};
  END {of movecursor};

PROCEDURE EditCommands;
  VAR
    Ch : CHAR;
  BEGIN {editcommands}
  Message('Delete,Insert,Copy,Edit name');
  GetChar(Ch,[#27,#32..#126] + MoveSet);
  CASE Ch OF
    'D' : DeleteOrInsert('D');
    'I' : DeleteOrInsert('I');
    'C' : CopyGulp;
    'E' : EditName;
    ELSE
      BEGIN
      IF Ch IN MoveSet+PromptSet THEN MoveCursor(Ch);
      END;
    END {of case};
  Message(Prompt);
  END {of editcommands};

PROCEDURE JumpToExtreme;
  VAR
    OldFirstRow : RowType;
    OldFirstCol : ColType;
    Ch : CHAR;
  BEGIN {tab}
  Message('Top,Bottom,Left,Right');
  OldFirstRow := FirstRow; OldFirstCol := FirstCol;
  GetChar(Ch,[#27,#32..#126] + MoveSet);
  CASE Ch OF
    'T' : MoveToExtreme(RowsPerPage,NoRows,0,1,R,SR,FirstRow,TRUE);
    'B' : MoveToExtreme(RowsPerPage,NoRows,NoRows,RowsPerPage,R,SR,FirstRow,TRUE);
    'L' : MoveToExtreme(ColsPerPage,NoCols,0,1,C,SC,FirstCol,FALSE);
    'R' : MoveToExtreme(ColsPerPage,NoCols,NoCols,ColsPerPage,C,SC,FirstCol,FALSE);
    ELSE
      BEGIN
      IF Ch IN MoveSet+PromptSet THEN MoveCursor(Ch);
      END;
    END {of case};
  Message(Prompt);
  END {jumptoextreme};

PROCEDURE Commands;
  VAR
    C : STRING[9];
  BEGIN
  IF It = -1 THEN C := 'Calculate' ELSE C := 'Continue';
  Message(C+',Help,Edit,File,Alter,Print,Mark,Jump,Quit');
  GetChar(Ch,[#27,#32..#126] + MoveSet);
  CASE Ch OF
    'C' : FinishEdit;
    'H' : Help;
    'E' : EditCommands;
    'F' : FileCommands(FALSE);
    'A' : Alter;
    'P' : Print;
    'M' : DefineBlock;
    'J' : JumpToExtreme;
    'Q' : Quit;
    '/' : PrintPage;
    ELSE
      BEGIN
      IF Ch IN MoveSet+PromptSet THEN MoveCursor(Ch);
      Message(Prompt);
      END;
    END {of case};
  END {of commands};

BEGIN {edit}
SetUpEditor;
REPEAT
  PrintPage;
  IF KEYPRESSED THEN Ch := ReadKey;
UNTIL NOT PrintPageStopped;
REPEAT
  OccupyCell;
  Response := READKEY;
  ConvertKeyToIBM(Response);
  IF Response = F1Key THEN Help
  ELSE IF Response IN MoveSet+PromptSet THEN MoveCursor(Response)
  ELSE IF Response IN ['!'..'.','0'..'~'] THEN GetCell(Response);
  IF PrintPageStopped THEN PrintPage;
UNTIL Ch = 'C';
END {of edit};

{$F+}
PROCEDURE CleanUpOnExit;
  VAR
    N : INTEGER;
  BEGIN
  ExitProc := ExitProcAddress;
  {$I-}
  N := LENGTH(GulpDir);
  IF (GulpDir[N] = '\') AND (N > 1) THEN
    IF (GulpDir[N-1] <> ':') THEN DELETE(GulpDir,N,1);
  ChDir(GULPDir);
  IF IORESULT <> 0 THEN BEGIN END;
  { $I+}
  TitleScreen;
  END;

FUNCTION HeapErrorFunc(Size:Word): INTEGER; far;
  BEGIN {HeapErrorFunc}
  IF Size > 0 THEN BEGIN
    ErrorMessage(1018);
    Halt;
    END;
  END {HeapErrorFunc};

BEGIN {main program}
ExitProcAddress := ExitProc;
ExitProc := @CleanUpOnExit;
HeapError := @HeapErrorFunc;
IF KEYPRESSED THEN Ch := READKEY;
InitialSetUp;
ReadDefaults;
TitleScreen;
IF (Mode <> 'Batch') {AND (PARAMCOUNT = 0)} THEN BEGIN
  GOTOXY(1,ScreenHeight-1);
  FOR I := 1 TO ScreenWidth DO WRITE(#196);
  GOTOXY(1,ScreenHeight);
  Centre('Press any key');
  I := 0;
  REPEAT
    I := I + 1;
    Delay(1);
  UNTIL KEYPRESSED {OR (I = 60000)};
  IF KEYPRESSED THEN Ch := READKEY;
  CLRSCR;
  END {if}
ELSE Delay(3000);
IF PARAMCOUNT = 0 THEN WorkFile := DirString+'DEFAULT.GLP';
Initialise(MaxNoRows,MaxNoCols);
REPEAT
  IF Exist(WorkFile) AND NOT (FirstTime AND (PARAMCOUNT = 0)) THEN
    BEGIN
    Message('Retrieving '+WorkFile);
    Load(WorkFile);
    END
  ELSE IF Mode = 'Batch' THEN
    BEGIN
    ErrorMessage(1021);
    HALT;
    END;
  IF Mode <> 'Batch' THEN Edit(Workfile);
  Solve;
  IF NOT JumpToEditor AND NOT StopSolve THEN PrintSoln(TRUE);
  IF Mode <> 'Batch' THEN Initialise(NoRows,NoCols);
UNTIL Mode = 'Batch';
END.
